@@ -42,7 +42,7 @@
 #include <EGL/eglext.h>
 
 #include "compositor.h"
-#include "../shared/cairo-util.h"
+#include "../shared/image-loader.h"
 
 struct wayland_compositor {
 	struct weston_compositor	 base;
